'' =================================================================================================
''
''   File....... jm_elock_timer.spin
''   Purpose.... Electronic Lock Demo for Propeller Activity Board 
''   Author..... Jon "JonnyMac" McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 20 JUL 2016
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal
                                                                 
  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us
                                                                  
                                                                  
con { pab io pins }                                               
                                                                  
  RX1     = 31                                                  ' programming / terminal
  TX1     = 30                                      
                                                    
  SDA1    = 29                                                  ' eeprom / i2c
  SCL1    = 28                                                   
                                                                 
  AUD_RT  = 27                                                  ' audio (d/a) outputs
  AUD_LF  = 26                                                   
                                                                 
  LED_27  = 27                                                  ' leds (on audio outputs)
  LED_26  = 26                                                   
                                                                 
  SD_CS   = 25                                                  ' usd card
  SD_DI   = 24                                                   
  SD_CLK  = 23                                                   
  SD_DO   = 22                                                   
                                                                 
  ADC_CS  = 21                                                  ' adc input
  ADC_CLK = 20                                                   
  ADC_DO  = 19                                                   
  ADC_DI  = 18                                                   


con { application io }

  UNLOCK  = 26                                                  ' indicate unlock state with LED 26
                                                         
  SDA2    = 15                                                  ' 4x14segs display connection
  SCL2    = 14

  COL3    =  7                                                  ' keypad columns
  COL2    =  6
  COL1    =  5
  COL0    =  4
  
  ROW3    =  3                                                  ' keypad rows
  ROW2    =  2
  ROW1    =  1
  ROW0    =  0                                        
                                                                 
                                                                 
con

  #0, G_LOCK, G_OPEN
                                                                 
  KEY_TIMEOUT = 3000 { ms }                                         
                                                                 
                                                                 
obj                                                              
                                                                 
' main                                                          ' * master spin cog
  time       : "jm_time"                                        '   delays/timing
  prng       : "jm_prng"                                        '   random #s
  io         : "jm_io"                                          '   basic io
  segs14     : "jm_ht16k33_4x14segs"                            '   4-digit, 14-segment driver
  segs07     : "jm_ht16k33_4x7segs"                             '   4-digit, 7-segment driver  
  ee         : "jm_24xx256"                                     '   eeprom access
  adc        : "jm_adc124s021"                                  '   for onboard adc (foreground)
  term       : "jm_fullduplexserial"                            ' * serial for terminal
  sd         : "fsrw"                                           ' * uSD file IO
' background                                                    ' * support processes
                                                           
' * uses cog when loaded                                         
                                                                 
                                                                 
var

  byte  keybuf[5]                                               ' 4 keys + null
  byte  codelen                                                 ' # keys in buffer

  byte  gstate                                                  ' game state

  long  ktimer                                                  ' key timeout timer


dat

  GameCode      byte    "1234", 0                               ' default unlock code
  GameTime      long    30_00                                   ' default game time (decimal mmss)                                                       
                                                                 
                                                                 
pub main                                     

  setup                                                         ' start objects

  check_update                                                  ' check for [*] + [#]
  
  repeat
    reset_game
    update_display                                              ' show state + time
    wait_key("#")
    start_game_timer      
    repeat
      update_display                                            ' show state/code + time
      if (grunning)                                             ' if still running
        check_user_key                                          '  look for a key input
      else
        time.pause(10_000)                                      ' no input for 10s
        wait_key("*")                                           ' wait for reset
        quit                                                    ' start again


pub reset_game

  set_game_timer(GameTime / 100, GameTime // 100)               ' set starting time
  segs07.set_blink(segs07#BLINK_OFF)                            ' solid
  gstate := G_LOCK                                              ' set state to locked 

  clear_code                                                    ' wipe out user code

  ktimer := millis - KEY_TIMEOUT                                ' force lock state display
  
  
pub clear_code
  
  bytemove(@keybuf, string("****"), 5)                          ' fill buffer with "empty" key slots
  codelen := 0


pub wait_key(target) | k

  flush_keypad
  repeat
    k := key_to_ascii(read_keypad)                              ' get a key
  until (k == target)                                           ' wait for #


pub update_display

  if ((millis - ktimer) < KEY_TIMEOUT)                          ' okay to show last key?
    segs14.str(@keybuf) 
  else                                                          ' key timeout, show lock state    
    if (gstate == G_LOCK)
      segs14.str(string("LOCK"))
      clear_code
    else
      segs14.str(string("OPEN"))

  segs07.clock(gmins, gsecs)                                    ' show game time

  if ((gmins == 0) and (gsecs == 0))                            ' game over?
    segs07.set_blink(segs07#BLINK_2HZ)                                                        


pub check_user_key | k

  ifnot (grunning)
    return

  k := key_to_ascii(read_keypad)

  case k
    "0".."9":
      if (codelen < 4)                                          ' room in current buffer?
        keybuf[codelen++] := k                                  ' yes, add key
        ktimer := millis                                        ' reset key timer 

    "*" :
      if (codelen > 0)                                          ' anything in buffer?
        keybuf[--codelen] := "*"                                ' yes, delete last
        ktimer := millis                                        ' reset key timer

    "#" :
      if (codelen == 4)                                         ' full buffer?
        if (strcomp(@GameCode, @keybuf))                        ' match for game code?
          stop_game_timer                                       ' yes! winner winner chicken dinner
          gstate := G_OPEN                                      ' change game state
        ktimer := millis - KEY_TIMEOUT                          ' force state display 


dat

  PromptCode    byte    "   ENTER UNLOCK CODE", 0
  PromptTime    byte    "   ENTER GAME TIME", 0


pub check_update | idx, gtime

  time.pause(KEY_MS << 1)                                       ' give time to scan keypad

  if (read_keypad <> STAR_POUND)                                ' reprogram request
    flush_keypad
    return
   
  segs14.scroll_str(@PromptCode, 250)  
  get_code(4)                                                   ' get 4-digit code

  bytemove(@GameCode, @keybuf, 4)                               ' update key code     
  repeat idx from 0 to 3                                        ' save to EEPROM      
    ee.wr_byte(@GameCode[idx], GameCode[idx])                                         

  time.pause(500)                                               ' hold just a bit
  segs07.clear(true)
    
  repeat
    segs14.scroll_str(@PromptTime, 250)
    gtime := get_time                                           ' get game time as decimal
    if (gtime > 0)                                              ' if time entered
      GameTime := gtime                                         '  update program
      ee.wr_long(@GameTime, GameTime)
      quit      
  

pub get_code(reqlen) | len, k

'' Get code from keypad
'' -- * is backspace, # is enter

  flush_keypad                                                  ' clear keypad buffer

  len := 0                                                      ' no keys yet

  bytemove(@keybuf, string("----"), 5) 
  segs07.str(@keybuf)                  
 
  repeat
    segs07.str(@keybuf)                                         ' show current code    

    k := key_to_ascii(read_keypad)                              ' get key, convert to ASCII
    case k
      "0".."9":                                                 ' if digit
        if (len < 4)                                            ' and not done         
          keybuf[len++] := k                                    ' add key to buffer                               

      "*" :                                                     ' backspace?
        if (len > 0)
          keybuf[--len] := "-"   
            
      "#" :                                                     ' enter?
        if (reqlen > 0)                                         ' required length set?
          if (len == reqlen)                                    ' match
            return len
        else
          return len


pub get_time | value, key

'' Get time as decimal value

  flush_keypad                                                  ' clear keypad buffer

  value := 0

  repeat
    segs07.dec(value)                                           ' update display
    segs07.show_colon(value => 100)                             ' light colon when minutes used
      
    key := key_to_ascii(read_keypad)                            ' get a key 
    case key
      "0".."9" :                                                ' if digit
        value *= 10                                             '  shift value left
        value += (key - "0")                                    '  add new 1s

      "*" :
        value /= 10                                             ' delete 1s

      "#" :                                                     ' return value to caller
        return value

                                                             
pub setup                                                        
                                                                 
  time.start                                                    ' setup for timing/delays
                                                                 
  prng.seed(-cnt, -cnt ~> 2, $DEAD_BEEF, cnt << 2, cnt)         ' seed randomizer                             
                                                                 
  io.start(0, 0)                                                ' clear pins (this cog)

  ee.start(%000)                                                ' connect to eeprom

  segs14.start(SCL2, SDA2, %000)                                ' connect to 4x14s display
  segs07.start(SCL2, SDA2, %001)                                ' connect to 4x7s display

' adc.start(ADC_CS, ADC_CLK, ADC_DI, ADC_DO)                    ' connect to adc 
                                                        
' term.start(RX1, TX1, %0000, 115_200)                          ' start serial for terminal *
                                                                 
' mount_sd                                                      ' start sd access *

  bcog := cognew(background, @bstack) + 1                       ' launch background cog *
                                                                 
                                                                 
con                                                              
                                                                 
  { ----------------- }                                          
  {  F I L E   I / O  }                                          
  { ----------------- }                                          
                                                                 
                                                                 
var                                                              
                                                                 
  long  mounted                                                  
                                                                 
  byte  filename[8 + 1 + 3 + 1]                                 ' 8.3 plus 0
  byte  linebuf[80 + 1]                                         ' line buffer
                                                                 
                                                                 
pub mount_sd | check                                             
                                                                 
'' Attempt to mount SD card                                      
'' -- returns true if SD card successfully mounted

  check := \sd.mount_explicit(SD_DO, SD_CLK, SD_DI, SD_CS)

  mounted := (check == 0)   

  return mounted


pub dismount_sd

'' Unmounts SD card

  \sd.unmount
  mounted := false


pub has_file(p_str) | check

'' Returns true if file is on SD card

  check := \sd.popen(p_str, "r")
  if (check == 0)
    return true
  else
    return false 


pub open_file(p_str, mode) | check

'' Attempt to open file
'' -- p_str is pointer to filename (z-string)
'' -- mode is [lower-case] single character, (e.g., "r" for read)
'' -- sd card must be mounted before open

  if (mounted)                                                   
    check := \sd.popen(p_str, mode)                             ' attempt to open
    if (check == 0)                                              
      if (mode == "r")                                           
        \sd.seek(0)                                             ' force sector read 
      return true                                                
    else                                                         
      return false
  else
    return false


pub close_file

'' Closes open file

  \sd.pclose


pub rd_line(p_str, n) | c, len

'' Reads line of text from open text file
'' -- terminated by LF or EOF
'' -- p_str is address of string buffer
'' -- n is maximum number of characters to read from line

  len := 0                                                      ' index into string
                                                                 
  repeat n                                                       
    c := \sd.pgetc                                              ' get a character
    if (c < 0)                                                  ' end of file
      if (len == 0)                                             ' if nothing
        return -1                                               ' return empty
      else                                                       
        quit                                                     
                                                                 
    if (c <> 10)                                                ' if not LF char
      byte[p_str++] := c                                        '  add to buffer
      len++                                                      
    else                                                        ' if LF
      quit                                                      '  we're done
                                                                 
  byte[p_str] := 0                                              ' terminate end of line
                                                                 
  return len
  

pub wr_line(p_str, usecr, uself)

'' Writes z-string to open text file
'' -- set usecr to true to add CR
'' -- set uself to true to add LF

  \sd.pputs(p_str)
  if (usecr)
    \sd.pputc(13)
  if (uself)
    \sd.pputc(10)


pub rd_buf(p_buf, n) | check

'' Reads n bytes from file to buffer at p_buf
'' -- returns number of bytes read

  return \sd.pread(p_buf, n)


pub wr_buf(p_buf, n) | check   

'' Writes n bytes to file from buffer at p_buf
'' -- returns number of bytes written

  return \sd.pwrite(p_buf, n)


pub directory | fcount, check, idx, p_fname, fsize

'' Show directory (root only) of uSD card

  sd.opendir                                                    ' setup for .nextfile
  fcount := 0                                                   ' reset files count
                                                                 
  term.str(string("Directory", term#CR, term#CR))                
                                                                 
  repeat                                                         
    bytefill(@filename, 0, 13)                                   
    check := \sd.nextfile(@filename)                            ' get next file in root
    if (check == 0)                                             ' if found
      fcount += 1                                               '  increment file count
      term.tx(" ")
      term.str(@filename)                                       '  show name
      term.tx(13)                                           
    else                                                        ' else
      quit                                                      '  done
                                                                 
  if (fcount > 0)                                                
    term.tx(13)
    term.dec(fcount)
    term.str(string(" file(s) found."))
  else
    term.str(string("No files found."))

  sd.pclose


pub show_file(p_file) | len

'' Display contents of text file

  ifnot (mounted)
    term.str(string("SD not mounted.", 13))
    return

  if (open_file(p_file, "r"))
    repeat
      bytefill(@linebuf, 0, 81)
      len := rd_line(@linebuf, 80)
      if (len > 0)
        term.str(@linebuf)
        time.pause(5)
      else
        close_file
        quit


con                                                              
                                                                 
  { --------------------- }                                          
  {  B A C K G R O U N D  }                                          
  { --------------------- }


pub blink_led(led, onms, offms)

'' Update LED parameters for background control
'' -- led is 26 or 27
'' -- on and off times in milliseconds

  if (led == LED_26)
    led26on := onms
    led26cycles := onms + offms
    led26phase := 0

  elseif (led == LED_27)
    led27on := onms
    led27cycles := onms + offms
    led27phase := 0


pub dim_led(led, duty)

'' Set selected LED to specified duty cycle
'' -- duty is 0..100%
'' -- period fixed at 50Hz

  if (led == LED_26)
    led26on := (duty * 200 / 100 + 5) / 10                      ' round up
    led26cycles := 20
    led26phase := 0

  elseif (led == LED_27)
    led27on := (duty * 200 / 100 + 5) / 10
    led27cycles := 20
    led27phase := 0


pub tone_on(pin, hz, ms)

'' Play tone on pin
'' -- frequency in hz
'' -- ms is duration in milliseconds

  tone_off                                                      ' kill current tone
  waitcnt(cnt + MS_001)                                         ' ensure it's off
  longmove(@tonepin, @pin, 3)                                   ' copy parameters


pub tone_off

'' Kill tone now

  tonems := 0


pub read_keypad

'' Returns next key in buffer
'' -- 0 if buffer empty

  if (ptail <> phead)                                           ' if key present
    result := padbuf[ptail++]                                   ' return next in buffer
    ptail &= %1111                                              ' update tail pointer

  else
    result := 0                                                 ' no key


pub flush_keypad

'' Clears keypad buffer and resets head & tail pointers

  longfill(@padbuf, 0, 18)


pub key_to_ascii(kmask)

'' Convert keyboard mask to ascii code

  case kmask
    KEY_1 : return "1"
    KEY_2 : return "2"   
    KEY_3 : return "3"     
    KEY_4 : return "4"   
    KEY_5 : return "5"   
    KEY_6 : return "6"     
    KEY_7 : return "7"  
    KEY_8 : return "8"  
    KEY_9 : return "9" 
    KEY_0 : return "0"
    KEY_A : return "A"                                          ' use ascii values for non-digits
    KEY_B : return "B"
    KEY_C : return "C"
    KEY_D : return "D" 
    KEY_S : return "*"                                         
    KEY_P : return "#"

  return -1


pub set_game_timer(gm, gs)

  longfill(@gmillis, 4, 0)                                      ' clear all
  gmins := gm // 100                                            ' keep legal for display
  gsecs := gs // 100


pub start_game_timer

  gmillis := 0
  grunning := true


pub stop_game_timer

  grunning := false


pub resume_game_timer

  if ((gsecs > 0) or (gmins > 0))
    grunning := true
                                 

var

  long  bcog                                                    ' cog running background code
  long  bstack[64]                                              ' stack space for background cog

  long  millis                                                  ' running milliseconds


pri background | t                                              ' launch with cognew()

  io.low(LED_26)                                                ' setup LED pins
  io.low(LED_27)

  io.high(ADC_CS)                                               ' setup ADC pins
  io.low(ADC_CLK)
  io.low(ADC_DI)
  io.input(ADC_DO)
  
  t := cnt                                                      ' sync loop timing
  repeat
    waitcnt(t += MS_001)                                        ' wait 1ms
    ++millis                                                    ' increment millis
    update_leds
    play_tone
    scan_keypad
    game_timer
    

var

  long  led26on
  long  led26cycles
  long  led26phase

  long  led27on
  long  led27cycles
  long  led27phase
  

pri update_leds                                                 ' call from background()

  outa[LED_26] := (led26phase < led26on)    
  if (++led26phase == led26cycles)
    led26phase := 0
  
  outa[LED_27] := (led27phase < led27on)  
  if (++led27phase == led27cycles)
    led27phase := 0
  

var

  long  tonepin
  long  tonefreq
  long  tonems


pri play_tone                                                   ' call from background()  

  if (tonems > 0)                                               ' if timer running
    if (ctra == 0)                                              ' tone not enabled yet?
      frqa := ($8000_0000 / (CLK_FREQ / tonefreq)) << 1         ' setup for frequency
      ctra := ((%00100) << 26) | tonepin                        ' activate tone on pin
      dira[tonepin] := 1                                        ' make pin an output
    else
      --tonems
  else
    ctra := 0                                                   ' stop tone
    outa[tonepin] := 0                                          ' clear pin


var

  long  padwork
  long  padkey                                                  ' last key read
  long  padcycles                                               ' cycles last key held 

  long  padbuf[16]                                              ' keep as contiguous group
  long  phead
  long  ptail


pri scan_keypad | row 

  padwork := 0

  repeat row from 0 to 3
    outa[ROW0..ROW3] := |<row                                   ' enable one column
    dira[ROW0..ROW3] := |<row 
    padwork |= ina[COL3..COL0] << ((3-row) << 2)                ' shift nib into position

  dira[ROW0..ROW3] := %0000                                     ' disable all row outputs
    
  if (padwork == padkey)                                        ' no change?                        
    if (++padcycles == KEY_MS)                                  ' debounced?             
      if (padkey)                                               ' key pressed?                      
        padbuf[phead++] := padkey                               ' add to buffer, advance index      
        phead &= %1111                                          ' roll-over index                   
    else                                                                                            
      if (padcycles == AUTO_KEY)                                                                    
        padbuf[phead++] := padkey                               ' add to buffer, advance index      
        phead &= %1111                                          ' roll-over index                   
        padcycles -= KEY_RATE                                                                       
                                                                                                    
  else                                                                                          
    padkey := padwork                                           ' new key mask                                                                        
    padcycles := 1                                              ' restart debounce timing                                   


con { keypad connections }

  {
                COL0   COL1   COL2   COL2 
                  |      |      |      |  
     ROW0 --1K---[1]----[2]----[3]----[A] 
                  |      |      |      |   
     ROW1 --1K---[4]----[5]----[6]----[B]  
                  |      |      |      |    
     ROW2 --1K---[7]----[8]----[9]----[C]   
                  |      |      |      |    
     ROW3 --1K---[*]----[0]----[#]----[D]
                  |      |      |      |
                 10K    10K    10K    10K
                  |      |      |      |
                 Gnd    Gnd    Gnd    Gnd

     Columns are pulled low through 10K resistors

  }


  ' set AUTO_KEY to negative value to disable

  KEY_MS    =    25                                             ' key debounce timing
  AUTO_KEY = -1000                                              ' if key held 1s
  KEY_RATE =   250                                              '  repeat every 1/2s
  

con { key masks }

' mask     D#0* C987 B654 A321
                              
  KEY_1 = %0000_0000_0000_0001
  KEY_2 = %0000_0000_0000_0010
  KEY_3 = %0000_0000_0000_0100  
  KEY_4 = %0000_0000_0001_0000
  KEY_5 = %0000_0000_0010_0000
  KEY_6 = %0000_0000_0100_0000
  KEY_7 = %0000_0001_0000_0000
  KEY_8 = %0000_0010_0000_0000
  KEY_9 = %0000_0100_0000_0000
  KEY_0 = %0010_0000_0000_0000

  KEY_A = %0000_0000_0000_1000  
  KEY_B = %0000_0000_1000_0000
  KEY_C = %0000_1000_0000_0000 
  KEY_D = %1000_0000_0000_0000
  
  KEY_S = %0001_0000_0000_0000                                  ' star *                       
  KEY_P = %0100_0000_0000_0000                                  ' pound #

  
  STAR_POUND = KEY_S | KEY_P                                    ' specialty key


var

  long  gmillis
  long  gsecs
  long  gmins
  long  grunning
  

pri game_timer

  if (grunning)
    if (++gmillis == 1000)                                      ' new second?
      gmillis := 0                                              ' yes, reset
      if (--gsecs < 0)                                          ' if roll-under
        if (gmins)                                              '  and minutes left
          gsecs := 59                                           ' update secs
          --gmins                                               ' decrement mins
        else
          longfill(@gsecs, 0, 3)                                ' stop timer


dat { license }

{{

  Copyright (c) 2016 Jon McPhalen  

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}