'' =================================================================================================
''
''   File....... jm_ht16k33.spin
''   Purpose.... Spin/I2C driver for Holtek HT16K33 16x8 segments driver
''               -- output only; no key scanning
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2014-2016 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 12 JUL 2016
''
'' =================================================================================================


con { fixed io pins }

  RX1 = 31                                                      ' programming / terminal
  TX1 = 30                                                       
                                                                 
  SDA = 29                                                      ' eeprom / i2c
  SCL = 28                                                       


con

  HT16K33 = %1110_000_0

  #0, DISP_OFF, DISP_ON 

  #0, BLINK_OFF, BLINK_2HZ, BLINK_1HZ, BLINK_0x5HZ


obj

  i2c : "jm_i2c"


var

  long  slaveid                                                 ' slave id for this device

  word  dispbuf[8]                                              ' support up to 8 digits (2 bytes each)

  byte  dispctrl                                                ' display control bits
  byte  posidx                                                  ' display position index
  

pub null

  ' This is not a top-level object   

            
pub start(dscl, dsda, device)

'' Starts i2c display device
'' -- dscl is the device SCL pin
'' -- dsda is the device SDA pin
'' -- device is address, %000 to %111

  i2c.setupx(dscl, dsda)                                        ' connect to display

  slaveid := HT16K33 | (device << 1)                            ' set slave id

  write_cmd($21)                                                ' turn on oscillator  
  
  clear(true)                                                   ' clear display buffer
  
  set_display(DISP_ON, BLINK_OFF)
  set_brightness(15)                                            ' max brightness


pub set_display(state, rate)

'' Update display control bits
'' -- state is off (0) or on (1)
'' -- rate is the blink rate (0..3)

  if (state)
    state := 1                                                  ' promote any true to 1, false to 0

  if ((rate < BLINK_OFF) or (rate > BLINK_0x5HZ))
    rate := BLINK_OFF

  dispctrl := $80 | (rate << 1) | state

  write_cmd(dispctrl)


pub display_on(state)

'' Sets display on or off

  if (state)
    dispctrl |= %0000_0001                                      ' display bit on
  else
    dispctrl &= %1111_1110                                      ' display bit off

  write_cmd(dispctrl)


pub set_blink(rate)

  if ((rate < BLINK_OFF) or (rate > BLINK_0x5HZ))               ' if bad rate
    rate := BLINK_OFF                                           '  set to none

  dispctrl &= %1111_0001                                        ' remove old blink bits
  dispctrl |= rate << 1                                         ' update rate

  write_cmd(dispctrl)
   

pub set_brightness(b)

'' Set display brightness, 0..15

  write_cmd($E0 | (0 #> b <# 15)) 


pub clear(update) 

'' Clears display buffer

  wordfill(@dispbuf, 0, 8)

  if (update)
    write_buf(0, @dispbuf, 16)


pub refresh(n) | idx

'' Writes display buffer to device
'' -- limit to n digits/characters

  write_buf(0, @dispbuf, n<<1)


pub put_buf(col, bits)

'' Updates display buffer position with bits

  if ((col < 0) or (col > 7))                                   ' validate position
    return -1

  dispbuf[col] := bits
  

pub direct_write(col, bits)

'' Writes directly to display
'' -- buffer is updated as well

  if (put_buf(col, bits) < 0)
    return

  col <<= 1

  write_buf(col, @dispbuf+col, 2) 


pub address

'' Returns hub address of display buffer

  return @dispbuf


pri write_cmd(cmd)

  i2c.wait(slaveid)
  i2c.write(cmd)
  i2c.stop


pri write_buf(addr, p_src, count)

  i2c.wait(slaveid)
  i2c.write(addr)
  repeat count
    i2c.write(byte[p_src++])  
  i2c.stop

  
dat { license }

{{

  Copyright (C) 2014-2016 Jon McPhalen 

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}