'' =================================================================================================
''
''   File....... jm_ht16k33_4x7segs.spin
''   Purpose.... 4x7 segment display driver (e.g., Adafruit HT16K33 backpack)
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2016 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 20 JUL 2016
''
'' =================================================================================================


con { fixed io pins }

  RX1 = 31                                                      ' programming / terminal
  TX1 = 30
  
  SDA = 29                                                      ' eeprom / i2c
  SCL = 28
  

con

  DISP_OFF    = segs#DISP_OFF
  DISP_ON     = segs#DISP_ON 

  BLINK_OFF   = segs#BLINK_OFF  
  BLINK_2HZ   = segs#BLINK_2HZ  
  BLINK_1HZ   = segs#BLINK_1HZ  
  BLINK_0x5HZ = segs#BLINK_0x5HZ

  BLANK       = %00000000
  DP_SEG      = %10000000
  COLON_SEG   = %00000010
  AM_SEG      = %00000100
  PM_SEG      = %00001000
  DEG_SEG     = %00010000


obj

  segs : "jm_ht16k33"                                           ' HT16K33 segment controller


var

  byte  dots                                                    ' colon, am/pm, degrees dots


pub null

  ' This is not a top-level object   

            
pub start(dscl, dsda, device)

'' Starts i2c display device
'' -- device is address, %000 to %111
'' -- dscl is the device SCL pin
'' -- dsda is the device SDA pin

  segs.start(dscl, dsda, device)


pub set_display(state, rate)

'' Update display control bits
'' -- state is off (0) or on (1)
'' -- rate is the blink rate (0..3)

  segs.set_display(state, rate) 


pub display_on(state)

'' Sets display on or off

  segs.display_on(state)  


pub set_blink(rate)

  segs.set_blink(rate)
   

pub set_brightness(b)

'' Set display brightness, 0..15

  segs.set_brightness(b)   


pub clear(update) 

'' Clears display buffer

  segs.clear(update)


pub refresh(n) | idx

'' Writes display buffer to device
'' -- limit to n digits/characters
'' -- n is 1..5

  segs.refresh(n)


pub put_buf(col, bits)

'' Updates display buffer position with bits
'' -- pos is 0..4
''    * col 4 is colon/dots 
'' -- bits is 8-bit (7 + dp) segments

  segs.put_buf(remap(col), bits)

  
pub put_all(bits)

'' Fills each column in display buffer with bits

  wordfill(segs.address, bits, 5)
  segs.put_buf(2, 0)                                            ' clear colon/dots reg


pub direct_write(col, bits)

'' Writes directly to display
'' -- buffer is updated as well
'' -- pos is 0..5
'' -- bits is 8-bit segments

  segs.direct_write(col, bits)     


pub alpha_bits(alpha)

'' Convert alpha to 7-segment bit pattern

  if (alpha => "0") and (alpha =< "9")
    return DecMap[alpha-"0"]

  else
    case alpha
      " "   : return %000_0000
      34    : return %010_0010                                  ' double quote
      39    : return %000_0010                                  ' single quote
      "("   : return %011_1001                                  ' uses ]
      ")"   : return %000_1111                                  ' uses [
      "-"   : return %100_0000
      "="   : return %100_1000
      "?"   : return %101_0011
      "A"   : return DIG_A 
      "C"   : return DIG_C
      "E"   : return DIG_E
      "F"   : return DIG_F 
      "G"   : return DIG_6
      "H"   : return %111_0110
      "I"   : return DIG_1   
      "J"   : return %001_1110
      "L"   : return %011_1000 
      "O"   : return DIG_0
      "P"   : return %111_0011
      "S"   : return DIG_5
      "U"   : return %011_1110
      "Y"   : return %110_1110
      "["   : return %011_1001
      "]"   : return %000_1111  
      "_"   : return %000_1000
      "b"   : return DIG_b
      "c"   : return %101_1000
      "d"   : return DIG_d
      "h"   : return %111_0100 
      "i"   : return %000_0100
      "l"   : return DIG_1
      "n"   : return %101_0100
      "o"   : return %101_1100
      "q"   : return %110_0111
      "r"   : return %101_0000
      "t"   : return %111_0000
      "u"   : return %001_1100
      other : return %000_0000

  
pub digit(d)

'' Converts decimal digit d to bit pattern

  if ((d => 0) and (d =< 9))
    return DecMap[d]


pub str(p_str) | last, col

'' Print string at p_str
'' -- if shorter than 4 characters, trailing spaces added

  last := (strsize(p_str) <# 4) - 1
  if (last < 0)
    return

  repeat col from 0 to 3
    if (col =< last)
      segs.put_buf(remap(col), alpha_bits(byte[p_str++]))       ' output character          
    else                                                                                    
      segs.put_buf(remap(col), 0)                               ' pad with trailing spaces  

  segs.refresh(5)                                               ' 4 chars/digits + colon

  
pub scroll_str(p_str, ms)

  ms *= clkfreq / 1000                                          ' convert millis to ticks

  repeat strsize(p_str)-3                                       ' scroll to end of string   
    str(p_str++)                                                ' output 4 chars      
    waitcnt(cnt + ms)                                           ' hold
  

pub bin(value) | col

  value &= %1111                                                ' truncate to 4 bits

  repeat col from 0 to 3
    if (value & |<(3-col))                                      ' flip lsb to right
      segs.put_buf(remap(col), DIG_1)
    else
      segs.put_buf(remap(col), DIG_0)

  segs.refresh(5)         


pub dec(value) | sign, divisor, col, dig

  segs.clear(false)                                             ' clear display buffer

  value := -999 #> value <# 9_999                               ' limit range of value

  if (value < 0)                                                ' check for - sign
    sign := "-"
    -value    
  else
    sign := 0

  if (value == 0)                                               ' zero is special case
    divisor := 1
    col := 3
  else
    divisor := 1000                                             ' setup for max width
    col := 0
    repeat                                                      ' correct divisor for value
      if (value => divisor)
        quit
      divisor /= 10
      col += 1

  if (sign)                                                     ' display sign if needed
    segs.put_buf(col-1, alpha_bits("-"))

  repeat while (col < 4)                                        ' put valu digits in buffer
    dig := value / divisor 
    segs.put_buf(remap(col++), DecMap[dig])
    value //= divisor
    divisor /= 10 
  
  segs.refresh(5)


pub hex(value) | col, dig

  value <<= 16
  
  repeat col from 0 to 3
    dig := (value <-= 4) & $F
    segs.put_buf(remap(col), DecMap[dig])
    
  segs.refresh(5)


pub clock(m, s)

'' Display valus in clock format (MM:SS)
'' -- field values: 0..99

  put_buf(0, DecMap[m  / 10]) 
  put_buf(1, DecMap[m // 10])
  put_buf(2, DecMap[s  / 10]) 
  put_buf(3, DecMap[s // 10])

  show_colon(true)

  segs.refresh(5)  


pub set_dpoints(dpbits) | p_buf, col

'' Set decimal points in display
'' -- %0001 is LSB dp, %1000 is msb dp

  p_buf := segs.address                                         ' base address of segments buffer

  repeat col from 0 to 3
    if (dpbits & |<(3-col))
      word[p_buf][remap(col)] |= DP_SEG
    else
      word[p_buf][remap(col)] &= !DP_SEG    

  segs.refresh(5)


pub show_dots(dotbits)                                          ' for 1.2" display

'' Controls dots in display
'' -- bit0 is center colon
'' -- bit1 is upper dot of left colon (am)
'' -- bit2 is lower dot of left colon (pm)
'' -- bit3 is degrees dot

  dots := (dotbits & %1111) << 1
  segs.direct_write(2, dots)


pub show_colon(state)

  if (state)
    dots |= COLON_SEG                                           ' enable
  else
    dots &= !COLON_SEG                                          ' disable  

  segs.direct_write(2, dots)


pub show_colon2(state)

  if (state)
    dots |= (AM_SEG | PM_SEG)
  else
    dots &= !(AM_SEG | PM_SEG)

  segs.direct_write(2, dots)


pub show_am(state)

  if (state)
    dots |= AM_SEG
  else
    dots &= !AM_SEG

  segs.direct_write(2, dots)


pub show_pm(state)

  if (state)
    dots |= PM_SEG
  else
    dots &= !PM_SEG

  segs.direct_write(2, dots)


pub show_degrees(state)

  if (state)
    dots |= DEG_SEG
  else
    dots &= !DEG_SEG

  segs.direct_write(2, dots)


pub address

'' Returns hub address of display buffer

  return segs.address


pri remap(c)

'' Remap virtual column to physical column in Adafruit display

  if ((c => 0) and (c =< 4))
    return lookupz(c : 0, 1, 3, 4, 2)
  else
    return c


con

'          .gfedcba
  DIG_0 = %00111111
  DIG_1 = %00000110
  DIG_2 = %01011011  
  DIG_3 = %01001111 
  DIG_4 = %01100110 
  DIG_5 = %01101101
  DIG_6 = %01111101
  DIG_7 = %00000111
  DIG_8 = %01111111
  DIG_9 = %01101111
  
  DIG_A = %01110111
  DIG_b = %01111100
  DIG_C = %00111001
  DIG_d = %01011110
  DIG_E = %01111001
  DIG_F = %01110001


dat

  DecMap                byte    DIG_0, DIG_1, DIG_2, DIG_3, DIG_4
                        byte    DIG_5, DIG_6, DIG_7, DIG_8, DIG_9
                        
  HexMap                byte    DIG_A, DIG_b, DIG_C, DIG_d, DIG_E, DIG_F


dat { license }

{{

  Copyright (C) 2016 Jon McPhalen 

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}